@@ -40,7 +40,6 @@ int set_essid(struct wrap_ndis_device *wnd, const char *ssid, int ssid_len)
 	NDIS_STATUS res;
 	struct ndis_essid req;
 
-	ENTER2("");
 	if (ssid_len > NDIS_ESSID_MAX_SIZE)
 		return -EINVAL;
 
@@ -419,8 +418,7 @@ static int iw_set_tx_power(struct net_device *dev, struct iw_request_info *info,
 					&ndis_power, sizeof(ndis_power));
 		if (res)
 			return -EOPNOTSUPP;
-		res = miniport_set_info(wnd, OID_802_11_DISASSOCIATE,
-					NULL, 0);
+		res = disassociate(wnd, 1);
 		if (res)
 			return -EOPNOTSUPP;
 		return 0;
@@ -674,13 +672,13 @@ int get_auth_mode(struct wrap_ndis_device *wnd)
 	ULONG mode;
 	NDIS_STATUS res;
 
-	ENTER2("");
 	res = miniport_query_int(wnd, OID_802_11_AUTHENTICATION_MODE, &mode);
 	if (res) {
 		WARNING("getting authentication mode failed (%08X)", res);
 		EXIT2(return -EOPNOTSUPP);
-	} else
-		EXIT2(return mode);
+	}
+	TRACE2("%d", mode);
+	return mode;
 }
 
 int set_encr_mode(struct wrap_ndis_device *wnd, ULONG encr_mode)
@@ -1352,7 +1350,7 @@ static int iw_get_ndis_stats(struct net_device *dev,
 			     union iwreq_data *wrqu, char *extra)
 {
 	struct wrap_ndis_device *wnd = netdev_priv(dev);
-	struct iw_statistics *stats = &wnd->wireless_stats;
+	struct iw_statistics *stats = &wnd->iw_stats;
 	memcpy(&wrqu->qual, &stats->qual, sizeof(stats->qual));
 	return 0;
 }
@@ -1472,17 +1470,23 @@ static int iw_get_range(struct net_device *dev, struct iw_request_info *info,
 	return 0;
 }
 
-static int disassociate(struct wrap_ndis_device *wnd)
+NDIS_STATUS disassociate(struct wrap_ndis_device *wnd, int reset_ssid)
 {
+	NDIS_STATUS res;
 	u8 buf[NDIS_ESSID_MAX_SIZE];
 	int i;
 
-	miniport_set_info(wnd, OID_802_11_DISASSOCIATE, NULL, 0);
-	get_random_bytes(buf, sizeof(buf));
-	for (i = 0; i < sizeof(buf); i++)
-		buf[i] = 'a' + (buf[i] % 26);
-	set_essid(wnd, buf, sizeof(buf));
-	return 0;
+	TRACE2("");
+	res = miniport_set_info(wnd, OID_802_11_DISASSOCIATE, NULL, 0);
+	/* disassociate causes radio to be turned off; if reset_ssid
+	 * is given, set ssid to random to enable radio */
+	if (reset_ssid) {
+		get_random_bytes(buf, sizeof(buf));
+		for (i = 0; i < sizeof(buf); i++)
+			buf[i] = 'a' + (buf[i] % 26);
+		set_essid(wnd, buf, sizeof(buf));
+	}
+	return res;
 }
 
 
@@ -1491,7 +1495,7 @@ static int deauthenticate(struct wrap_ndis_device *wnd)
 	int ret;
 
 	ENTER2("");
-	ret = disassociate(wnd);
+	ret = disassociate(wnd, 1);
 	set_priv_filter(wnd, Ndis802_11PrivFilterAcceptAll);
 	set_auth_mode(wnd, Ndis802_11AuthModeOpen);
 	set_encr_mode(wnd, Ndis802_11EncryptionDisabled);
@@ -1523,7 +1527,7 @@ static int iw_set_mlme(struct net_device *dev, struct iw_request_info *info,
 		return deauthenticate(wnd);
 	case IW_MLME_DISASSOC:
 		TRACE2("cmd=%d reason_code=%d", mlme->cmd, mlme->reason_code);
-		return disassociate(wnd);
+		return disassociate(wnd, 1);
 	default:
 		return -EOPNOTSUPP;
 	}
@@ -1887,7 +1891,7 @@ static int priv_network_type(struct net_device *dev,
 		network_type = Ndis802_11DS;
 	else if (type == 'a')
 		network_type = Ndis802_11OFDM5;
-	else if (type == 'g')
+	else if (type == 'g' || type == 'n')
 		network_type = Ndis802_11OFDM24;
 	else
 		network_type = Ndis802_11Automode;
@@ -2132,7 +2136,7 @@ static int wpa_disassociate(struct net_device *dev,
 			    union iwreq_data *wrqu, char *extra)
 {
 	struct wrap_ndis_device *wnd = netdev_priv(dev);
-	disassociate(wnd);
+	disassociate(wnd, 1);
 	EXIT2(return 0);
 }
 
@@ -2394,6 +2398,6 @@ const struct iw_handler_def ndis_handler_def = {
 	.private	= (iw_handler *)priv_handler,
 	.private_args	= (struct iw_priv_args *)priv_args,
 #if WIRELESS_EXT >= 19
-	.get_wireless_stats = get_wireless_stats,
+	.get_wireless_stats = get_iw_stats,
 #endif
 };
