@@ -150,12 +150,24 @@ static int procfs_read_ndis_hw(char *page, char **start, off_t off,
 	ndis_fragmentation_threshold frag_threshold;
 	ndis_antenna antenna;
 	ULONG packet_filter;
+	int n;
+	mac_address mac;
+	char *hw_status[] = {"ready", "initializing", "resetting", "closing",
+			     "not ready"};
 
 	if (off != 0) {
 		*eof = 1;
 		return 0;
 	}
 
+	res = miniport_query_int(wnd, OID_GEN_HARDWARE_STATUS, &n);
+	if (res >= 0 && res < sizeof(hw_status) / sizeof(hw_status[0]))
+		p += sprintf(p, "status=%s\n", hw_status[res]);
+
+	res = miniport_query_info(wnd, OID_802_3_CURRENT_ADDRESS,
+				  mac, sizeof(mac));
+	if (!res)
+		p += sprintf(p, "mac: " MACSTRSEP "\n", MAC2STR(mac));
 	res = miniport_query_info(wnd, OID_802_11_CONFIGURATION,
 				  &config, sizeof(config));
 	if (!res) {
@@ -338,9 +350,9 @@ static int procfs_write_ndis_settings(struct file *file, const char *buf,
 		p++;
 		i = simple_strtol(p, NULL, 10);
 		if (i > 0)
-			wnd->stats_enabled = TRUE;
+			wnd->iw_stats_enabled = TRUE;
 		else
-			wnd->stats_enabled = FALSE;
+			wnd->iw_stats_enabled = FALSE;
 	} else if (!strcmp(setting, "packet_filter")) {
 		if (!p)
 			return -EINVAL;
@@ -376,7 +388,7 @@ int wrap_procfs_add_ndis_device(struct wrap_ndis_device *wnd)
 					 wnd->procfs_iface);
 	if (procfs_entry == NULL) {
 		ERROR("couldn't create proc entry for 'hw'");
-		return -ENOMEM;
+		goto err_hw;
 	} else {
 		procfs_entry->uid = proc_uid;
 		procfs_entry->gid = proc_gid;
@@ -388,7 +400,7 @@ int wrap_procfs_add_ndis_device(struct wrap_ndis_device *wnd)
 					 wnd->procfs_iface);
 	if (procfs_entry == NULL) {
 		ERROR("couldn't create proc entry for 'stats'");
-		return -ENOMEM;
+		goto err_stats;
 	} else {
 		procfs_entry->uid = proc_uid;
 		procfs_entry->gid = proc_gid;
@@ -400,7 +412,7 @@ int wrap_procfs_add_ndis_device(struct wrap_ndis_device *wnd)
 					 wnd->procfs_iface);
 	if (procfs_entry == NULL) {
 		ERROR("couldn't create proc entry for 'encr'");
-		return -ENOMEM;
+		goto err_encr;
 	} else {
 		procfs_entry->uid = proc_uid;
 		procfs_entry->gid = proc_gid;
@@ -413,7 +425,7 @@ int wrap_procfs_add_ndis_device(struct wrap_ndis_device *wnd)
 					 S_IWUSR | S_IWGRP, wnd->procfs_iface);
 	if (procfs_entry == NULL) {
 		ERROR("couldn't create proc entry for 'settings'");
-		return -ENOMEM;
+		goto err_settings;
 	} else {
 		procfs_entry->uid = proc_uid;
 		procfs_entry->gid = proc_gid;
@@ -422,11 +434,22 @@ int wrap_procfs_add_ndis_device(struct wrap_ndis_device *wnd)
 		procfs_entry->write_proc = procfs_write_ndis_settings;
 	}
 	return 0;
+
+err_settings:
+	remove_proc_entry("encr", wnd->procfs_iface);
+err_encr:
+	remove_proc_entry("stats", wnd->procfs_iface);
+err_stats:
+	remove_proc_entry("hw", wnd->procfs_iface);
+err_hw:
+	remove_proc_entry(wnd->netdev_name, wrap_procfs_entry);
+	wnd->procfs_iface = NULL;
+	return -ENOMEM;
 }
 
 void wrap_procfs_remove_ndis_device(struct wrap_ndis_device *wnd)
 {
-	struct proc_dir_entry *procfs_iface = wnd->procfs_iface;
+	struct proc_dir_entry *procfs_iface = xchg(&wnd->procfs_iface, NULL);
 
 	if (procfs_iface == NULL)
 		return;
@@ -434,9 +457,8 @@ void wrap_procfs_remove_ndis_device(struct wrap_ndis_device *wnd)
 	remove_proc_entry("stats", procfs_iface);
 	remove_proc_entry("encr", procfs_iface);
 	remove_proc_entry("settings", procfs_iface);
-	if (wrap_procfs_entry != NULL)
+	if (wrap_procfs_entry)
 		remove_proc_entry(wnd->netdev_name, wrap_procfs_entry);
-	wnd->procfs_iface = NULL;
 }
 
 static int procfs_read_debug(char *page, char **start, off_t off,
