@@ -20,8 +20,13 @@
 #define FALSE				0
 
 #define PASSIVE_LEVEL			0
+#define APC_LEVEL			1
 #define DISPATCH_LEVEL			2
-#define DEVICE_LEVEL			(DISPATCH_LEVEL + 1)
+#define DEVICE_LEVEL_BASE		4
+
+/* soft interrupts / bottom-half's are disabled at SOFT_IRQL */
+#define SOFT_IRQL			(DEVICE_LEVEL_BASE + 1)
+#define DIRQL				(DEVICE_LEVEL_BASE + 2)
 
 #define STATUS_WAIT_0			0
 #define STATUS_SUCCESS                  0
@@ -94,9 +99,9 @@
 #define THREAD_WAIT_OBJECTS		3
 #define MAX_WAIT_OBJECTS		64
 
-#define LOW_PRIORITY			1
+#define LOW_PRIORITY			0
 #define LOW_REALTIME_PRIORITY		16
-#define HIGH_PRIORITY			32
+#define HIGH_PRIORITY			31
 #define MAXIMUM_PRIORITY		32
 
 #define PROCESSOR_FEATURE_MAX		64
@@ -228,7 +233,9 @@ struct nt_list {
 	struct nt_list *prev;
 };
 
-typedef ULONG_PTR NT_SPIN_LOCK;
+typedef volatile ULONG_PTR NT_SPIN_LOCK;
+
+enum kdpc_importance {LowImportance, MediumImportance, HighImportance};
 
 struct kdpc;
 typedef void (*DPC)(struct kdpc *kdpc, void *ctx, void *arg1,
@@ -242,7 +249,12 @@ struct kdpc {
 	void *ctx;
 	void *arg1;
 	void *arg2;
-	NT_SPIN_LOCK *lock;
+	union {
+		NT_SPIN_LOCK *lock;
+		/* 'lock' is not used; 'queued' represents whether
+		 * kdpc is queued or not */
+		int queued;
+	};
 };
 
 enum pool_type {
@@ -360,7 +372,7 @@ struct wait_block {
 	struct nt_list list;
 	struct task_struct *thread;
 	void *object;
-	void *thread_waitq;
+	int *wait_done;
 	USHORT wait_key;
 	USHORT wait_type;
 };
@@ -417,7 +429,7 @@ struct nt_timer {
 		ULONGLONG due_time;
 		struct wrap_timer *wrap_timer;
 	};
-	struct nt_list list;
+	struct nt_list nt_timer_list;
 	struct kdpc *kdpc;
 	union {
 		LONG period;
@@ -444,15 +456,18 @@ struct nt_thread {
 	 * structure is opaque to drivers, we just define what we
 	 * need */
 	int pid;
+	NTSTATUS status;
 	struct task_struct *task;
 	struct nt_list irps;
 	NT_SPIN_LOCK lock;
 };
 
-#define set_dh_type(dh, type)		((dh)->type = (type))
-#define is_mutex_dh(dh)			((dh)->type == MutexObject)
-#define is_semaphore_dh(dh)		((dh)->type == SemaphoreObject)
-#define is_nt_thread_dh(dh)		((dh)->type == ThreadObject)
+#define set_object_type(dh, type)	((dh)->type = (type))
+#define is_notify_object(dh)		((dh)->type == NotificationObject)
+#define is_synch_object(dh)		((dh)->type == SynchronizationObject)
+#define is_mutex_object(dh)		((dh)->type == MutexObject)
+#define is_semaphore_object(dh)		((dh)->type == SemaphoreObject)
+#define is_nt_thread_object(dh)		((dh)->type == ThreadObject)
 
 #define IO_TYPE_ADAPTER				1
 #define IO_TYPE_CONTROLLER			2
@@ -1011,7 +1026,7 @@ struct irp {
 		LONG irp_count;
 		void *system_buffer;
 	} associated_irp;
-	struct nt_list threads;
+	struct nt_list thread_list;
 	struct io_status_block io_status;
 	KPROCESSOR_MODE requestor_mode;
 	BOOLEAN pending_returned;
@@ -1076,14 +1091,20 @@ struct irp {
 
 #define IoSetNextIrpStackLocation(irp)				\
 do {								\
+	KIRQL _irql_;						\
+	IoAcquireCancelSpinLock(&_irql_);			\
 	(irp)->current_location--;				\
 	IoGetCurrentIrpStackLocation(irp)--;			\
+	IoReleaseCancelSpinLock(_irql_);			\
 } while (0)
 
 #define IoSkipCurrentIrpStackLocation(irp) 			\
 do {								\
+	KIRQL _irql_;						\
+	IoAcquireCancelSpinLock(&_irql_);			\
 	(irp)->current_location++;				\
 	IoGetCurrentIrpStackLocation(irp)++;			\
+	IoReleaseCancelSpinLock(_irql_);			\
 } while (0)
 
 static inline void
@@ -1150,15 +1171,6 @@ enum key_value_information_class {
 	KeyValuePartialInformationAlign64
 };
 
-struct object_attr {
-	ULONG length;
-	void *root_dir;
-	struct unicode_string *name;
-	ULONG attr;
-	void *security_descriptor;
-	void *security_qos;
-};
-
 struct file_name_info {
 	ULONG length;
 	wchar_t *name;
@@ -1359,6 +1371,10 @@ struct callback_object {
 	struct object_attributes *attributes;
 };
 
+enum section_inherit {
+	ViewShare = 1, ViewUnmap = 2
+};
+
 struct ksystem_time {
 	ULONG low_part;
 	LONG high1_time;
