@@ -29,25 +29,21 @@ extern int hangcheck_interval;
 extern struct iw_handler_def ndis_handler_def;
 extern NT_SPIN_LOCK timer_lock;
 
+/* use own workqueue instead of shared one, to avoid depriving
+ * others */
+workqueue_struct_t *wrapndis_wq;
+static struct nt_thread *wrapndis_worker_thread;
+
 static int set_packet_filter(struct wrap_ndis_device *wnd,
 			     ULONG packet_filter);
-static void add_stats_timer(struct wrap_ndis_device *wnd);
-static void del_stats_timer(struct wrap_ndis_device *wnd);
-static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd);
-static int ndis_remove_device(struct wrap_ndis_device *wnd);
+static void add_iw_stats_timer(struct wrap_ndis_device *wnd);
+static void del_iw_stats_timer(struct wrap_ndis_device *wnd);
+static NDIS_STATUS wrap_ndis_start_device(struct wrap_ndis_device *wnd);
+static int wrap_ndis_remove_device(struct wrap_ndis_device *wnd);
 static void set_multicast_list(struct wrap_ndis_device *wnd);
 static int ndis_net_dev_open(struct net_device *net_dev);
 static int ndis_net_dev_close(struct net_device *net_dev);
 
-static inline int ndis_wait_comm_completion(struct wrap_ndis_device *wnd)
-{
-	if ((wait_event_interruptible(wnd->ndis_comm_wq,
-				      (wnd->ndis_comm_done > 0))))
-		return -1;
-	else
-		return 0;
-}
-
 /* MiniportReset */
 NDIS_STATUS miniport_reset(struct wrap_ndis_device *wnd)
 {
@@ -58,17 +54,17 @@ NDIS_STATUS miniport_reset(struct wrap_ndis_device *wnd)
 	KIRQL irql;
 
 	ENTER2("wnd: %p", wnd);
-
 	if (down_interruptible(&wnd->tx_ring_mutex))
 		EXIT3(return NDIS_STATUS_FAILURE);
 	if (down_interruptible(&wnd->ndis_comm_mutex)) {
 		up(&wnd->tx_ring_mutex);
 		EXIT3(return NDIS_STATUS_FAILURE);
-	}		
+	}
 	miniport = &wnd->wd->driver->ndis_driver->miniport;
 	cur_lookahead = wnd->nmb->cur_lookahead;
 	max_lookahead = wnd->nmb->max_lookahead;
 	wnd->ndis_comm_done = 0;
+	wnd->ndis_comm_task = current;
 	WARNING("%s is being reset", wnd->net_dev->name);
 	irql = serialize_lock_irql(wnd);
 	res = LIN2WIN2(miniport->reset, &reset_address, wnd->nmb->adapter_ctx);
@@ -77,7 +73,8 @@ NDIS_STATUS miniport_reset(struct wrap_ndis_device *wnd)
 	TRACE2("%08X, %08X", res, reset_address);
 	if (res == NDIS_STATUS_PENDING) {
 		/* wait for NdisMResetComplete */
-		if (ndis_wait_comm_completion(wnd))
+		if (wait_condition((wnd->ndis_comm_done > 0), 0,
+				   TASK_INTERRUPTIBLE) < 0)
 			res = NDIS_STATUS_FAILURE;
 		else {
 			res = wnd->ndis_comm_status;
@@ -106,13 +103,12 @@ NDIS_STATUS miniport_query_info_needed(struct wrap_ndis_device *wnd,
 	struct miniport_char *miniport;
 	KIRQL irql;
 
-	ENTER2("oid: %08X", oid);
-
 	if (down_interruptible(&wnd->ndis_comm_mutex))
 		EXIT3(return NDIS_STATUS_FAILURE);
 	miniport = &wnd->wd->driver->ndis_driver->miniport;
 	TRACE2("%p, %08X", miniport->query, oid);
 	wnd->ndis_comm_done = 0;
+	wnd->ndis_comm_task = current;
 	irql = serialize_lock_irql(wnd);
 	res = LIN2WIN6(miniport->query, wnd->nmb->adapter_ctx, oid, buf,
 		       bufsize, &written, needed);
@@ -121,7 +117,8 @@ NDIS_STATUS miniport_query_info_needed(struct wrap_ndis_device *wnd,
 	TRACE2("%08X, %08X", res, oid);
 	if (res == NDIS_STATUS_PENDING) {
 		/* wait for NdisMQueryInformationComplete */
-		if (ndis_wait_comm_completion(wnd))
+		if (wait_condition((wnd->ndis_comm_done > 0), 0,
+				   TASK_INTERRUPTIBLE) < 0)
 			res = NDIS_STATUS_FAILURE;
 		else
 			res = wnd->ndis_comm_status;
@@ -155,13 +152,12 @@ NDIS_STATUS miniport_set_info(struct wrap_ndis_device *wnd, ndis_oid oid,
 	struct miniport_char *miniport;
 	KIRQL irql;
 
-	ENTER2("oid: %08X", oid);
-
 	if (down_interruptible(&wnd->ndis_comm_mutex))
 		EXIT3(return NDIS_STATUS_FAILURE);
 	miniport = &wnd->wd->driver->ndis_driver->miniport;
 	TRACE2("%p, %08X", miniport->query, oid);
 	wnd->ndis_comm_done = 0;
+	wnd->ndis_comm_task = current;
 	irql = serialize_lock_irql(wnd);
 	res = LIN2WIN6(miniport->setinfo, wnd->nmb->adapter_ctx, oid,
 		       buf, bufsize, &written, &needed);
@@ -170,7 +166,8 @@ NDIS_STATUS miniport_set_info(struct wrap_ndis_device *wnd, ndis_oid oid,
 	TRACE2("%08X, %08X", res, oid);
 	if (res == NDIS_STATUS_PENDING) {
 		/* wait for NdisMQueryInformationComplete */
-		if (ndis_wait_comm_completion(wnd))
+		if (wait_condition((wnd->ndis_comm_done > 0), 0,
+				   TASK_INTERRUPTIBLE) < 0)
 			res = NDIS_STATUS_FAILURE;
 		else
 			res = wnd->ndis_comm_status;
@@ -213,7 +210,7 @@ static NDIS_STATUS miniport_pnp_event(struct wrap_ndis_device *wnd,
 	}
 	/* RNDIS driver doesn't like to be notified if device is
 	 * already halted */
-	if (!test_bit(HW_INITIALIZED, &wnd->hw_status))
+	if (!test_bit(HW_INITIALIZED, &wnd->wd->hw_status))
 		EXIT1(return NDIS_STATUS_SUCCESS);
 	switch (event) {
 	case NdisDevicePnPEventSurpriseRemoved:
@@ -221,7 +218,7 @@ static NDIS_STATUS miniport_pnp_event(struct wrap_ndis_device *wnd,
 		       (wnd->attributes & NDIS_ATTRIBUTE_SURPRISE_REMOVE_OK),
 		       miniport->pnp_event_notify);
 		if ((wnd->attributes & NDIS_ATTRIBUTE_SURPRISE_REMOVE_OK) &&
-		    wnd->wd->surprise_removed == TRUE &&
+		    !test_bit(HW_PRESENT, &wnd->wd->hw_status) &&
 		    miniport->pnp_event_notify) {
 			TRACE1("calling surprise_removed");
 			LIN2WIN4(miniport->pnp_event_notify,
@@ -249,13 +246,12 @@ static NDIS_STATUS miniport_pnp_event(struct wrap_ndis_device *wnd,
 static NDIS_STATUS miniport_init(struct wrap_ndis_device *wnd)
 {
 	NDIS_STATUS error_status, status;
-	UINT medium_index;
-	UINT medium_array[] = {NdisMedium802_3};
+	UINT medium_index, medium_array[] = {NdisMedium802_3};
 	struct miniport_char *miniport;
 	struct ndis_pnp_capabilities pnp_capa;
 
 	ENTER1("irql: %d", current_irql());
-	if (test_bit(HW_INITIALIZED, &wnd->hw_status)) {
+	if (test_bit(HW_INITIALIZED, &wnd->wd->hw_status)) {
 		WARNING("device %p already initialized!", wnd);
 		return NDIS_STATUS_FAILURE;
 	}
@@ -278,12 +274,12 @@ static NDIS_STATUS miniport_init(struct wrap_ndis_device *wnd)
 
 	/* Wait a little to let card power up otherwise ifup might
 	 * fail after boot */
-	sleep_hz(HZ / 2);
+	sleep_hz(HZ / 5);
 	status = miniport_pnp_event(wnd, NdisDevicePnPEventPowerProfileChanged,
 				    NdisPowerProfileAcOnLine);
 	if (status != NDIS_STATUS_SUCCESS)
 		TRACE1("setting power failed: %08X", status);
-	set_bit(HW_INITIALIZED, &wnd->hw_status);
+	set_bit(HW_INITIALIZED, &wnd->wd->hw_status);
 	/* the description about NDIS_ATTRIBUTE_NO_HALT_ON_SUSPEND is
 	 * misleading/confusing */
 	status = miniport_query_info(wnd, OID_PNP_CAPABILITIES,
@@ -306,21 +302,21 @@ static void miniport_halt(struct wrap_ndis_device *wnd)
 	struct miniport_char *miniport;
 
 	ENTER1("%p", wnd);
-	if (test_and_clear_bit(HW_INITIALIZED, &wnd->hw_status)) {
+	if (test_and_clear_bit(HW_INITIALIZED, &wnd->wd->hw_status)) {
 		hangcheck_del(wnd);
-		del_stats_timer(wnd);
+		del_iw_stats_timer(wnd);
 		miniport = &wnd->wd->driver->ndis_driver->miniport;
 		TRACE1("halt: %p", miniport->miniport_halt);
 		LIN2WIN1(miniport->miniport_halt, wnd->nmb->adapter_ctx);
 		/* cancel any timers left by bugyy windows driver; also free
 		 * the memory for timers */
 		while (1) {
-			KIRQL irql;
 			struct nt_list *ent;
 			struct wrap_timer *wrap_timer;
+			KIRQL irql;
 
 			irql = nt_spin_lock_irql(&timer_lock, DISPATCH_LEVEL);
-			ent = RemoveHeadList(&wnd->timer_list);
+			ent = RemoveHeadList(&wnd->wrap_timer_list);
 			nt_spin_unlock_irql(&timer_lock, irql);
 			if (!ent)
 				break;
@@ -331,9 +327,9 @@ static void miniport_halt(struct wrap_ndis_device *wnd)
 			 * already */
 			if (del_timer_sync(&wrap_timer->timer))
 				WARNING("Buggy Windows driver left timer %p "
-					"running", &wrap_timer->timer);
+					"running", wrap_timer->nt_timer);
 			memset(wrap_timer, 0, sizeof(*wrap_timer));
-			slack_kfree(wrap_timer);
+			kfree(wrap_timer);
 		}
 	} else
 		WARNING("device %p is not initialized - not halting", wnd);
@@ -349,13 +345,14 @@ static NDIS_STATUS miniport_set_power_state(struct wrap_ndis_device *wnd,
 	if (state == NdisDeviceStateD0) {
 		status = NDIS_STATUS_SUCCESS;
 		up(&wnd->ndis_comm_mutex);
-		if (test_and_clear_bit(HW_HALTED, &wnd->hw_status)) {
+		if (test_and_clear_bit(HW_HALTED, &wnd->wd->hw_status)) {
 			status = miniport_init(wnd);
 			if (status == NDIS_STATUS_SUCCESS) {
 				set_packet_filter(wnd, wnd->packet_filter);
 				set_multicast_list(wnd);
 			}
-		} else if (test_and_clear_bit(HW_SUSPENDED, &wnd->hw_status)) {
+		} else if (test_and_clear_bit(HW_SUSPENDED,
+					      &wnd->wd->hw_status)) {
 			status = miniport_set_int(wnd, OID_PNP_SET_POWER,
 						  state);
 			if (status != NDIS_STATUS_SUCCESS)
@@ -372,7 +369,7 @@ static NDIS_STATUS miniport_set_power_state(struct wrap_ndis_device *wnd,
 			up(&wnd->tx_ring_mutex);
 			netif_device_attach(wnd->net_dev);
 			hangcheck_add(wnd);
-			add_stats_timer(wnd);
+			add_iw_stats_timer(wnd);
 			set_scan(wnd);
 		} else
 			WARNING("%s: couldn't set power to state %d; device not"
@@ -383,27 +380,26 @@ static NDIS_STATUS miniport_set_power_state(struct wrap_ndis_device *wnd,
 			EXIT1(return NDIS_STATUS_FAILURE);
 		netif_device_detach(wnd->net_dev);
 		hangcheck_del(wnd);
-		del_stats_timer(wnd);
+		del_iw_stats_timer(wnd);
 		status = NDIS_STATUS_NOT_SUPPORTED;
 		if (wnd->attributes & NDIS_ATTRIBUTE_NO_HALT_ON_SUSPEND) {
-			enum ndis_power_state pm_state = state;
 			if (wnd->ndis_wolopts) {
 				status =
 					miniport_set_int(wnd,
 							 OID_PNP_ENABLE_WAKE_UP,
 							 wnd->ndis_wolopts);
-				if (status == NDIS_STATUS_SUCCESS) {
-					if (wrap_is_pci_bus(wnd->wd->dev_bus))
-						pci_enable_wake(wnd->wd->pci.pdev,
-								PCI_D0, 1);
-				} else
+				if (status == NDIS_STATUS_SUCCESS &&
+				    wrap_is_pci_bus(wnd->wd->dev_bus))
+					pci_enable_wake(wnd->wd->pci.pdev,
+							PCI_D0, 1);
+				else
 					WARNING("%s: couldn't enable WOL: %08x",
 						wnd->net_dev->name, status);
 			}
 			status = miniport_set_int(wnd, OID_PNP_SET_POWER,
-						  pm_state);
+						  state);
 			if (status == NDIS_STATUS_SUCCESS)
-				set_bit(HW_SUSPENDED, &wnd->hw_status);
+				set_bit(HW_SUSPENDED, &wnd->wd->hw_status);
 			else
 				WARNING("suspend failed: %08X", status);
 		}
@@ -411,7 +407,7 @@ static NDIS_STATUS miniport_set_power_state(struct wrap_ndis_device *wnd,
 			WARNING("%s does not support power management; "
 				"halting the device", wnd->net_dev->name);
 			miniport_halt(wnd);
-			set_bit(HW_HALTED, &wnd->hw_status);
+			set_bit(HW_HALTED, &wnd->wd->hw_status);
 			status = STATUS_SUCCESS;
 		}
 		if (down_interruptible(&wnd->ndis_comm_mutex))
@@ -469,7 +465,7 @@ static int setup_tx_sg_list(struct wrap_ndis_device *wnd, struct sk_buff *skb,
 			    struct ndis_packet_oob_data *oob_data)
 {
 	struct ndis_sg_element *sg_element;
-	struct ndis_sg_list *tx_sg_list;
+	struct ndis_sg_list *sg_list;
 	int i;
 
 	ENTER3("%p, %d", skb, skb_shinfo(skb)->nr_frags);
@@ -485,14 +481,14 @@ static int setup_tx_sg_list(struct wrap_ndis_device *wnd, struct sk_buff *skb,
 		TRACE3("%Lx, %u", sg_element->address, sg_element->length);
 		return 0;
 	}
-	tx_sg_list = kmalloc(sizeof(*tx_sg_list) +
-			     (skb_shinfo(skb)->nr_frags + 1) *
-			     sizeof(*sg_element), GFP_ATOMIC);
-	if (!tx_sg_list)
+	sg_list = kmalloc(sizeof(*sg_list) +
+			  (skb_shinfo(skb)->nr_frags + 1) * sizeof(*sg_element),
+			  GFP_ATOMIC);
+	if (!sg_list)
 		return -ENOMEM;
-	tx_sg_list->nent = skb_shinfo(skb)->nr_frags + 1;
-	TRACE3("%p, %d", tx_sg_list, tx_sg_list->nent);
-	sg_element = tx_sg_list->elements;
+	sg_list->nent = skb_shinfo(skb)->nr_frags + 1;
+	TRACE3("%p, %d", sg_list, sg_list->nent);
+	sg_element = sg_list->elements;
 	sg_element->length = skb_headlen(skb);
 	sg_element->address =
 		PCI_DMA_MAP_SINGLE(wnd->wd->pci.pdev, skb->data,
@@ -507,7 +503,7 @@ static int setup_tx_sg_list(struct wrap_ndis_device *wnd, struct sk_buff *skb,
 				     PCI_DMA_TODEVICE);
 		TRACE3("%Lx, %u", sg_element->address, sg_element->length);
 	}
-	oob_data->ext.info[ScatterGatherListPacketInfo] = tx_sg_list;
+	oob_data->ext.info[ScatterGatherListPacketInfo] = sg_list;
 	return 0;
 }
 
@@ -516,21 +512,21 @@ static void free_tx_sg_list(struct wrap_ndis_device *wnd,
 {
 	int i;
 	struct ndis_sg_element *sg_element;
-	struct ndis_sg_list *tx_sg_list =
+	struct ndis_sg_list *sg_list =
 		oob_data->ext.info[ScatterGatherListPacketInfo];
-	sg_element = tx_sg_list->elements;
-	TRACE3("%p, %d", tx_sg_list, tx_sg_list->nent);
+	sg_element = sg_list->elements;
+	TRACE3("%p, %d", sg_list, sg_list->nent);
 	PCI_DMA_UNMAP_SINGLE(wnd->wd->pci.pdev, sg_element->address,
 			     sg_element->length, PCI_DMA_TODEVICE);
-	if (tx_sg_list->nent == 1)
+	if (sg_list->nent == 1)
 		EXIT3(return);
-	for (i = 1; i < tx_sg_list->nent; i++, sg_element++) {
+	for (i = 1; i < sg_list->nent; i++, sg_element++) {
 		TRACE3("%Lx, %u", sg_element->address, sg_element->length);
 		pci_unmap_page(wnd->wd->pci.pdev, sg_element->address,
 			       sg_element->length, PCI_DMA_TODEVICE);
 	}
-	TRACE3("%p", tx_sg_list);
-	kfree(tx_sg_list);
+	TRACE3("%p", sg_list);
+	kfree(sg_list);
 }
 
 static struct ndis_packet *alloc_tx_packet(struct wrap_ndis_device *wnd,
@@ -564,7 +560,6 @@ static struct ndis_packet *alloc_tx_packet(struct wrap_ndis_device *wnd,
 			return NULL;
 		}
 	}
-	TRACE2("%d, %d", wnd->tx_csum.value, skb->ip_summed);
 	if (skb->ip_summed == CHECKSUM_PARTIAL) {
 		struct ndis_tcp_ip_checksum_packet_info csum;
 		struct iphdr *ip = skb->nh.iph;
@@ -576,7 +571,6 @@ static struct ndis_packet *alloc_tx_packet(struct wrap_ndis_device *wnd,
 		else if (ip->protocol == IPPROTO_UDP)
 			csum.tx.udp = 1;
 //		csum->tx.ip = 1;
-		TRACE3("0x%05x", csum.value);
 		packet->private.flags |= NDIS_PROTOCOL_ID_TCP_IP;
 		oob_data->ext.info[TcpIpChecksumPacketInfo] =
 			(void *)(ULONG_PTR)csum.value;
@@ -596,11 +590,11 @@ void free_tx_packet(struct wrap_ndis_device *wnd, struct ndis_packet *packet,
 
 	ENTER3("%p, %08X", packet, status);
 	if (status == NDIS_STATUS_SUCCESS) {
-		pre_atomic_add(wnd->stats.tx_bytes, packet->private.len);
-		atomic_inc_var(wnd->stats.tx_packets);
+		pre_atomic_add(wnd->net_stats.tx_bytes, packet->private.len);
+		atomic_inc_var(wnd->net_stats.tx_packets);
 	} else {
 		TRACE1("packet dropped: %08X", status);
-		atomic_inc_var(wnd->stats.tx_dropped);
+		atomic_inc_var(wnd->net_stats.tx_dropped);
 	}
 	oob_data = NDIS_PACKET_OOB_DATA(packet);
 	if (wnd->sg_dma_size)
@@ -622,7 +616,6 @@ static u8 miniport_tx_packets(struct wrap_ndis_device *wnd, u8 start, u8 n)
 	struct miniport_char *miniport;
 	struct ndis_packet *packet;
 	u8 sent;
-	KIRQL irql;
 
 	TRACE3("%d, %d", start, n);
 	miniport = &wnd->wd->driver->ndis_driver->miniport;
@@ -632,12 +625,10 @@ static u8 miniport_tx_packets(struct wrap_ndis_device *wnd, u8 start, u8 n)
 				 &wnd->tx_ring[start], n);
 			sent = n;
 		} else {
-			irql = raise_irql(DISPATCH_LEVEL);
 			serialize_lock(wnd);
 			LIN2WIN3(miniport->send_packets, wnd->nmb->adapter_ctx,
 				 &wnd->tx_ring[start], n);
 			serialize_unlock(wnd);
-			lower_irql(irql);
 			for (sent = 0; sent < n && wnd->tx_ok; sent++) {
 				struct ndis_packet_oob_data *oob_data;
 				packet = wnd->tx_ring[start + sent];
@@ -679,10 +670,10 @@ static u8 miniport_tx_packets(struct wrap_ndis_device *wnd, u8 start, u8 n)
 			packet = wnd->tx_ring[start + sent];
 			oob_data = NDIS_PACKET_OOB_DATA(packet);
 			oob_data->status = NDIS_STATUS_NOT_RECOGNIZED;
-			irql = serialize_lock_irql(wnd);
+			if_serialize_lock(wnd);
 			res = LIN2WIN3(miniport->send, wnd->nmb->adapter_ctx,
 				       packet, packet->private.flags);
-			serialize_unlock_irql(wnd, irql);
+			if_serialize_unlock(wnd);
 			switch (res) {
 			case NDIS_STATUS_SUCCESS:
 				free_tx_packet(wnd, packet, res);
@@ -712,27 +703,32 @@ static void tx_worker(worker_param_t param)
 {
 	struct wrap_ndis_device *wnd;
 	s8 n;
+	KIRQL irql;
 
 	wnd = worker_param_data(param, struct wrap_ndis_device, tx_work);
 	ENTER3("tx_ok %d", wnd->tx_ok);
 	while (wnd->tx_ok) {
 		if (down_interruptible(&wnd->tx_ring_mutex))
 			break;
-		/* end == start if either ring is empty or full; in
-		 * the latter case is_tx_ring_full is set */
+		nt_spin_lock_bh(&wnd->tx_ring_lock);
 		n = wnd->tx_ring_end - wnd->tx_ring_start;
 		TRACE3("%d, %d, %d", wnd->tx_ring_start, wnd->tx_ring_end, n);
+		/* end == start if either ring is empty or full; in
+		 * the latter case is_tx_ring_full is set */
 		if (n == 0) {
 			if (wnd->is_tx_ring_full)
 				n = TX_RING_SIZE - wnd->tx_ring_start;
 			else {
+				nt_spin_unlock_bh(&wnd->tx_ring_lock);
 				up(&wnd->tx_ring_mutex);
 				break;
 			}
 		} else if (n < 0)
 			n = TX_RING_SIZE - wnd->tx_ring_start;
+		nt_spin_unlock_bh(&wnd->tx_ring_lock);
 		if (unlikely(n > wnd->max_tx_packets))
 			n = wnd->max_tx_packets;
+		irql = raise_irql(DISPATCH_LEVEL);
 		n = miniport_tx_packets(wnd, wnd->tx_ring_start, n);
 		if (n > 0) {
 			wnd->net_dev->trans_start = jiffies;
@@ -742,6 +738,7 @@ static void tx_worker(worker_param_t param)
 			if (netif_queue_stopped(wnd->net_dev))
 				netif_wake_queue(wnd->net_dev);
 		}
+		lower_irql(irql);
 		up(&wnd->tx_ring_mutex);
 		TRACE3("%d, %d, %d", wnd->tx_ring_start, wnd->tx_ring_end, n);
 	}
@@ -758,9 +755,7 @@ static int tx_skbuff(struct sk_buff *skb, struct net_device *dev)
 		WARNING("couldn't allocate packet");
 		return NETDEV_TX_BUSY;
 	}
-	/* no need for lock here - already called holding
-	 * net_dev->xmit_lock and tx_ring_end is not updated
-	 * elsewhere */
+	nt_spin_lock(&wnd->tx_ring_lock);
 	wnd->tx_ring[wnd->tx_ring_end++] = packet;
 	if (wnd->tx_ring_end == TX_RING_SIZE)
 		wnd->tx_ring_end = 0;
@@ -768,8 +763,9 @@ static int tx_skbuff(struct sk_buff *skb, struct net_device *dev)
 		wnd->is_tx_ring_full = 1;
 		netif_stop_queue(wnd->net_dev);
 	}
+	nt_spin_unlock(&wnd->tx_ring_lock);
 	TRACE3("ring: %d, %d", wnd->tx_ring_start, wnd->tx_ring_end);
-	schedule_wrap_work(&wnd->tx_work);
+	schedule_wrapndis_work(&wnd->tx_work);
 	return NETDEV_TX_OK;
 }
 
@@ -833,14 +829,15 @@ static int ndis_net_dev_open(struct net_device *net_dev)
 	}
 	netif_wake_queue(net_dev);
 	netif_poll_enable(net_dev);
-	return 0;
+	EXIT1(return 0);
 }
 
 static int ndis_net_dev_close(struct net_device *net_dev)
 {
+	ENTER1("%p", netdev_priv(net_dev));
 	netif_poll_disable(net_dev);
 	netif_tx_disable(net_dev);
-	return 0;
+	EXIT1(return 0);
 }
 
 static int ndis_change_mtu(struct net_device *net_dev, int mtu)
@@ -878,7 +875,7 @@ static void ndis_poll_controller(struct net_device *dev)
 static struct net_device_stats *ndis_get_stats(struct net_device *dev)
 {
 	struct wrap_ndis_device *wnd = netdev_priv(dev);
-	return &wnd->stats;
+	return &wnd->net_stats;
 }
 
 /* called from BH context */
@@ -886,26 +883,26 @@ static void ndis_set_multicast_list(struct net_device *dev)
 {
 	struct wrap_ndis_device *wnd = netdev_priv(dev);
 	set_bit(SET_MULTICAST_LIST, &wnd->wrap_ndis_pending_work);
-	schedule_wrap_work(&wnd->wrap_ndis_work);
+	schedule_wrapndis_work(&wnd->wrap_ndis_work);
 }
 
 /* called from BH context */
-struct iw_statistics *get_wireless_stats(struct net_device *dev)
+struct iw_statistics *get_iw_stats(struct net_device *dev)
 {
 	struct wrap_ndis_device *wnd = netdev_priv(dev);
-	return &wnd->wireless_stats;
+	return &wnd->iw_stats;
 }
 
-static void update_wireless_stats(struct wrap_ndis_device *wnd)
+static void update_iw_stats(struct wrap_ndis_device *wnd)
 {
-	struct iw_statistics *iw_stats = &wnd->wireless_stats;
+	struct iw_statistics *iw_stats = &wnd->iw_stats;
 	struct ndis_wireless_stats ndis_stats;
 	NDIS_STATUS res;
 	ndis_rssi rssi;
 	int qual;
 
 	ENTER2("%p", wnd);
-	if (wnd->stats_enabled == FALSE || !netif_carrier_ok(wnd->net_dev)) {
+	if (wnd->iw_stats_enabled == FALSE || !netif_carrier_ok(wnd->net_dev)) {
 		memset(iw_stats, 0, sizeof(*iw_stats));
 		EXIT2(return);
 	}
@@ -1096,53 +1093,45 @@ static void link_status_handler(struct wrap_ndis_device *wnd)
 	EXIT2(return);
 }
 
-static void stats_timer_proc(unsigned long data)
+static void iw_stats_timer_proc(unsigned long data)
 {
 	struct wrap_ndis_device *wnd = (struct wrap_ndis_device *)data;
 
-	ENTER2("");
-	if (wnd->stats_interval <= 0)
-		EXIT2(return);
-	set_bit(COLLECT_STATS, &wnd->wrap_ndis_pending_work);
-	schedule_wrap_work(&wnd->wrap_ndis_work);
-	mod_timer(&wnd->stats_timer, jiffies + wnd->stats_interval);
+	ENTER2("%d", wnd->iw_stats_interval);
+	if (wnd->iw_stats_interval > 0) {
+		set_bit(COLLECT_IW_STATS, &wnd->wrap_ndis_pending_work);
+		schedule_wrapndis_work(&wnd->wrap_ndis_work);
+	}
+	mod_timer(&wnd->iw_stats_timer, jiffies + wnd->iw_stats_interval);
 }
 
-static void add_stats_timer(struct wrap_ndis_device *wnd)
+static void add_iw_stats_timer(struct wrap_ndis_device *wnd)
 {
 	if (wnd->physical_medium != NdisPhysicalMediumWirelessLan)
 		return;
-	if (wnd->stats_interval < 0)
-		wnd->stats_interval *= -1;
-	wnd->stats_timer.data = (unsigned long)wnd;
-	wnd->stats_timer.function = stats_timer_proc;
-	mod_timer(&wnd->stats_timer, jiffies + wnd->stats_interval);
+	if (wnd->iw_stats_interval < 0)
+		wnd->iw_stats_interval *= -1;
+	wnd->iw_stats_timer.data = (unsigned long)wnd;
+	wnd->iw_stats_timer.function = iw_stats_timer_proc;
+	mod_timer(&wnd->iw_stats_timer, jiffies + wnd->iw_stats_interval);
 }
 
-static void del_stats_timer(struct wrap_ndis_device *wnd)
+static void del_iw_stats_timer(struct wrap_ndis_device *wnd)
 {
-	ENTER2("");
-	wnd->stats_interval *= -1;
-	del_timer_sync(&wnd->stats_timer);
+	ENTER2("%d", wnd->iw_stats_interval);
+	wnd->iw_stats_interval *= -1;
+	del_timer_sync(&wnd->iw_stats_timer);
 	EXIT2(return);
 }
 
 static void hangcheck_proc(unsigned long data)
 {
 	struct wrap_ndis_device *wnd = (struct wrap_ndis_device *)data;
-	BOOLEAN reset;
-	struct miniport_char *miniport;
 
-	ENTER2("");
-	if (wnd->hangcheck_interval <= 0)
-		EXIT2(return);
-	miniport = &wnd->wd->driver->ndis_driver->miniport;
-	if_serialize_lock(wnd);
-	reset = LIN2WIN1(miniport->hangcheck, wnd->nmb->adapter_ctx);
-	if_serialize_unlock(wnd);
-	if (reset) {
-		set_bit(MINIPORT_RESET, &wnd->wrap_ndis_pending_work);
-		schedule_wrap_work(&wnd->wrap_ndis_work);
+	ENTER3("%d", wnd->hangcheck_interval);
+	if (wnd->hangcheck_interval > 0) {
+		set_bit(HANGCHECK, &wnd->wrap_ndis_pending_work);
+		schedule_wrapndis_work(&wnd->wrap_ndis_work);
 	}
 	mod_timer(&wnd->hangcheck_timer, jiffies + wnd->hangcheck_interval);
 	EXIT3(return);
@@ -1166,7 +1155,7 @@ void hangcheck_add(struct wrap_ndis_device *wnd)
 
 void hangcheck_del(struct wrap_ndis_device *wnd)
 {
-	ENTER2("");
+	ENTER2("%d", wnd->hangcheck_interval);
 	if (wnd->hangcheck_interval > 0)
 		wnd->hangcheck_interval *= -1;
 	del_timer_sync(&wnd->hangcheck_timer);
@@ -1179,25 +1168,36 @@ static void wrap_ndis_worker(worker_param_t param)
 	struct wrap_ndis_device *wnd;
 
 	wnd = worker_param_data(param, struct wrap_ndis_device, wrap_ndis_work);
-
-	TRACE2("%lu", wnd->wrap_ndis_pending_work);
+	WORKTRACE("0x%lx", wnd->wrap_ndis_pending_work);
 
 	if (test_bit(SHUTDOWN, &wnd->wrap_ndis_pending_work))
-		EXIT3(return);
+		WORKEXIT(return);
 
 	if (test_and_clear_bit(SET_MULTICAST_LIST, &wnd->wrap_ndis_pending_work))
 		set_multicast_list(wnd);
 
-	if (test_and_clear_bit(COLLECT_STATS, &wnd->wrap_ndis_pending_work))
-		update_wireless_stats(wnd);
+	if (test_and_clear_bit(COLLECT_IW_STATS, &wnd->wrap_ndis_pending_work))
+		update_iw_stats(wnd);
 
 	if (test_and_clear_bit(LINK_STATUS_CHANGED,
 			       &wnd->wrap_ndis_pending_work))
 		link_status_handler(wnd);
 
-	if (test_and_clear_bit(MINIPORT_RESET, &wnd->wrap_ndis_pending_work))
-		miniport_reset(wnd);
-	EXIT3(return);
+	if (test_and_clear_bit(HANGCHECK, &wnd->wrap_ndis_pending_work)) {
+		struct miniport_char *miniport;
+		BOOLEAN reset;
+		KIRQL irql;
+
+		miniport = &wnd->wd->driver->ndis_driver->miniport;
+		irql = serialize_lock_irql(wnd);
+		reset = LIN2WIN1(miniport->hangcheck, wnd->nmb->adapter_ctx);
+		serialize_unlock_irql(wnd, irql);
+		if (reset) {
+			TRACE2("%s needs reset", wnd->net_dev->name);
+			miniport_reset(wnd);
+		}
+	}
+	WORKEXIT(return);
 }
 
 NDIS_STATUS ndis_reinit(struct wrap_ndis_device *wnd)
@@ -1218,17 +1218,37 @@ NDIS_STATUS ndis_reinit(struct wrap_ndis_device *wnd)
 	return status;
 }
 
-static void get_encryption_capa(struct wrap_ndis_device *wnd)
+static void get_encryption_capa(struct wrap_ndis_device *wnd, char *buf,
+				const int buf_len)
 {
 	int i, mode;
 	NDIS_STATUS res;
 	struct ndis_assoc_info ndis_assoc_info;
 	struct ndis_add_key ndis_key;
 	struct ndis_capability *c;
-	char *buf;
-	const int buf_len = 512;
 
 	ENTER1("%p", wnd);
+	/* set network type to g, b, or a, in that order */
+	res = miniport_query_info(wnd, OID_802_11_NETWORK_TYPES_SUPPORTED,
+				  buf, buf_len);
+	if (res == NDIS_STATUS_SUCCESS) {
+		struct network_type_list *net_types;
+		unsigned long types = 0;
+		net_types = (typeof(net_types))buf;
+		for (i = 0; i < net_types->num; i++) {
+			TRACE2("%d", net_types->types[i]);
+			set_bit(net_types->types[i], &types);
+		}
+		if (types & Ndis802_11OFDM24)
+			mode = Ndis802_11OFDM24;
+		else if (types & Ndis802_11DS)
+			mode = Ndis802_11DS;
+		else if (types & Ndis802_11OFDM5)
+			mode = Ndis802_11OFDM5;
+		else
+			mode = Ndis802_11DS;
+		miniport_set_int(wnd, OID_802_11_NETWORK_TYPE_IN_USE, mode);
+	}
 	/* check if WEP is supported */
 	if (set_encr_mode(wnd, Ndis802_11Encryption1Enabled) == 0 &&
 	    get_encr_mode(wnd) == Ndis802_11Encryption1KeyAbsent)
@@ -1292,19 +1312,11 @@ static void get_encryption_capa(struct wrap_ndis_device *wnd)
 	 * WPA */
 	set_bit(Ndis802_11AuthModeWPA, &wnd->capa.auth);
 
-	/* check for wpa2 */
-	buf = kmalloc(buf_len, GFP_KERNEL);
-	if (!buf) {
-		ERROR("couldn't allocate memory");
-		EXIT1(return);
-	}
 	memset(buf, 0, buf_len);
 	c = (struct ndis_capability *)buf;
 	res = miniport_query_info(wnd, OID_802_11_CAPABILITY, buf, buf_len);
-	if (!(res == NDIS_STATUS_SUCCESS && c->version == 2)) {
-		kfree(buf);
+	if (!(res == NDIS_STATUS_SUCCESS && c->version == 2))
 		EXIT1(return);
-	}
 	wnd->num_pmkids = c->num_PMKIDs;
 
 	for (i = 0; i < c->num_auth_encr_pair; i++) {
@@ -1339,7 +1351,6 @@ static void get_encryption_capa(struct wrap_ndis_device *wnd)
 			break;
 		}
 	}
-	kfree(buf);
 	EXIT1(return);
 }
 
@@ -1441,7 +1452,7 @@ wstdcall NTSTATUS NdisDispatchPnp(struct device_object *fdo, struct irp *irp)
 		status = IoSyncForwardIrp(pdo, irp);
 		if (status != STATUS_SUCCESS)
 			break;
-		if (ndis_start_device(wnd) == NDIS_STATUS_SUCCESS)
+		if (wrap_ndis_start_device(wnd) == NDIS_STATUS_SUCCESS)
 			status = STATUS_SUCCESS;
 		else
 			status = STATUS_FAILURE;
@@ -1460,7 +1471,7 @@ wstdcall NTSTATUS NdisDispatchPnp(struct device_object *fdo, struct irp *irp)
 	case IRP_MN_REMOVE_DEVICE:
 		TRACE1("%s", wnd->net_dev->name);
 		miniport_pnp_event(wnd, NdisDevicePnPEventSurpriseRemoved, 0);
-		if (ndis_remove_device(wnd)) {
+		if (wrap_ndis_remove_device(wnd)) {
 			status = STATUS_FAILURE;
 			break;
 		}
@@ -1742,8 +1753,7 @@ static int notifier_event(struct notifier_block *notifier, unsigned long event,
 
 	wnd = netdev_priv(net_dev);
 	/* called with rtnl lock held, so no need to lock */
-	switch (event) {
-	case NETDEV_CHANGENAME:
+	if (event == NETDEV_CHANGENAME) {
 		wrap_procfs_remove_ndis_device(wnd);
 		printk(KERN_INFO "%s: changing interface name from '%s' to "
 		       "'%s'\n", DRIVER_NAME, wnd->netdev_name, net_dev->name);
@@ -1751,11 +1761,8 @@ static int notifier_event(struct notifier_block *notifier, unsigned long event,
 		       sizeof(wnd->netdev_name));
 		wrap_procfs_add_ndis_device(wnd);
 		return NOTIFY_OK;
-		break;
-	default:
-		TRACE2("%lx", event);
-		break;
 	}
+	TRACE2("%s: %lx", wnd->netdev_name, event);
 	return NOTIFY_DONE;
 }
 
@@ -1763,35 +1770,35 @@ static struct notifier_block netdev_notifier = {
 	.notifier_call = notifier_event,
 };
 
-static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
+static NDIS_STATUS wrap_ndis_start_device(struct wrap_ndis_device *wnd)
 {
 	struct wrap_device *wd;
 	struct net_device *net_dev;
-	NDIS_STATUS ndis_status;
+	NDIS_STATUS status;
 	char *buf;
-	const int buf_size = 256;
+	const int buf_len = 256;
 	mac_address mac;
-	struct transport_header_offset transport_header_offset;
+	struct transport_header_offset *tx_header_offset;
 	int n;
 
-	ndis_status = miniport_init(wnd);
-	if (ndis_status == NDIS_STATUS_NOT_RECOGNIZED)
+	status = miniport_init(wnd);
+	if (status == NDIS_STATUS_NOT_RECOGNIZED)
 		EXIT1(return NDIS_STATUS_SUCCESS);
-	if (ndis_status != NDIS_STATUS_SUCCESS)
-		EXIT1(return ndis_status);
+	if (status != NDIS_STATUS_SUCCESS)
+		EXIT1(return status);
 	wd = wnd->wd;
 	net_dev = wnd->net_dev;
 
-	ndis_status = miniport_query_info(wnd, OID_802_3_CURRENT_ADDRESS,
-					  mac, sizeof(mac));
-	if (ndis_status) {
-		ERROR("couldn't get mac address: %08X", ndis_status);
+	status = miniport_query_info(wnd, OID_802_3_CURRENT_ADDRESS,
+				     mac, sizeof(mac));
+	if (status) {
+		ERROR("couldn't get mac address: %08X", status);
 		goto err_start;
 	}
 	TRACE1("mac:" MACSTRSEP, MAC2STR(mac));
-	ndis_status = miniport_query_int(wnd, OID_GEN_PHYSICAL_MEDIUM,
-					 &wnd->physical_medium);
-	if (ndis_status != NDIS_STATUS_SUCCESS)
+	status = miniport_query_int(wnd, OID_GEN_PHYSICAL_MEDIUM,
+				    &wnd->physical_medium);
+	if (status != NDIS_STATUS_SUCCESS)
 		wnd->physical_medium = NdisPhysicalMediumUnspecified;
 
 	get_supported_oids(wnd);
@@ -1809,7 +1816,7 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 	net_dev->do_ioctl = NULL;
 	if (wnd->physical_medium == NdisPhysicalMediumWirelessLan) {
 #if WIRELESS_EXT < 19
-		net_dev->get_wireless_stats = get_wireless_stats;
+		net_dev->get_wireless_stats = get_iw_stats;
 #endif
 		net_dev->wireless_handlers = &ndis_handler_def;
 	}
@@ -1818,13 +1825,13 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 #if defined(HAVE_ETHTOOL)
 	net_dev->ethtool_ops = &ndis_ethtool_ops;
 #endif
-	if (wnd->ndis_irq)
-		net_dev->irq = wnd->ndis_irq->irq.irq;
+	if (wnd->mp_interrupt)
+		net_dev->irq = wnd->mp_interrupt->irq;
 	net_dev->mem_start = wnd->mem_start;
 	net_dev->mem_end = wnd->mem_end;
-	ndis_status = miniport_query_int(wnd, OID_802_3_MAXIMUM_LIST_SIZE,
+	status = miniport_query_int(wnd, OID_802_3_MAXIMUM_LIST_SIZE,
 					 &wnd->multicast_size);
-	if (ndis_status != NDIS_STATUS_SUCCESS || wnd->multicast_size < 0)
+	if (status != NDIS_STATUS_SUCCESS || wnd->multicast_size < 0)
 		wnd->multicast_size = 0;
 	if (wnd->multicast_size > 0)
 		net_dev->flags |= IFF_MULTICAST;
@@ -1834,25 +1841,27 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 	net_dev->poll_controller = ndis_poll_controller;
 #endif
 
-	buf = kmalloc(buf_size, GFP_KERNEL);
+	buf = kmalloc(buf_len, GFP_KERNEL);
 	if (!buf) {
 		WARNING("couldn't allocate memory");
 		goto err_start;
 	}
 
-	set_task_offload(wnd, buf, buf_size);
+	set_task_offload(wnd, buf, buf_len);
+#ifdef NETIF_F_LLTX
+	net_dev->features |= NETIF_F_LLTX;
+#endif
 
 	if (register_netdev(net_dev)) {
 		ERROR("cannot register net device %s", net_dev->name);
 		goto err_register;
 	}
-	register_netdevice_notifier(&netdev_notifier);
 	memcpy(wnd->netdev_name, net_dev->name, sizeof(wnd->netdev_name));
-	memset(buf, 0, buf_size);
-	ndis_status = miniport_query_info(wnd, OID_GEN_VENDOR_DESCRIPTION,
-					  buf, buf_size);
-	if (ndis_status != NDIS_STATUS_SUCCESS) {
-		WARNING("couldn't get vendor information: 0x%x", ndis_status);
+	memset(buf, 0, buf_len);
+	status = miniport_query_info(wnd, OID_GEN_VENDOR_DESCRIPTION,
+				     buf, buf_len);
+	if (status != NDIS_STATUS_SUCCESS) {
+		WARNING("couldn't get vendor information: 0x%x", status);
 		buf[0] = 0;
 	}
 	wnd->drv_ndis_version = n = 0;
@@ -1872,25 +1881,24 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 		 * keep max at TX_RING_SIZE */
 		wnd->max_tx_packets = TX_RING_SIZE;
 	} else {
-		ndis_status =
-			miniport_query_int(wnd, OID_GEN_MAXIMUM_SEND_PACKETS,
-					   &wnd->max_tx_packets);
-		if (ndis_status != NDIS_STATUS_SUCCESS)
+		status = miniport_query_int(wnd, OID_GEN_MAXIMUM_SEND_PACKETS,
+					    &wnd->max_tx_packets);
+		if (status != NDIS_STATUS_SUCCESS)
 			wnd->max_tx_packets = 1;
 		if (wnd->max_tx_packets > TX_RING_SIZE)
 			wnd->max_tx_packets = TX_RING_SIZE;
 	}
 	TRACE2("maximum send packets: %d", wnd->max_tx_packets);
-	NdisAllocatePacketPoolEx(&ndis_status, &wnd->tx_packet_pool,
+	NdisAllocatePacketPoolEx(&status, &wnd->tx_packet_pool,
 				 wnd->max_tx_packets, 0,
 				 PROTOCOL_RESERVED_SIZE_IN_PACKET);
-	if (ndis_status != NDIS_STATUS_SUCCESS) {
+	if (status != NDIS_STATUS_SUCCESS) {
 		ERROR("couldn't allocate packet pool");
 		goto packet_pool_err;
 	}
-	NdisAllocateBufferPool(&ndis_status, &wnd->tx_buffer_pool,
+	NdisAllocateBufferPool(&status, &wnd->tx_buffer_pool,
 			       wnd->max_tx_packets + 4);
-	if (ndis_status != NDIS_STATUS_SUCCESS) {
+	if (status != NDIS_STATUS_SUCCESS) {
 		ERROR("couldn't allocate buffer pool");
 		goto buffer_pool_err;
 	}
@@ -1904,19 +1912,17 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 	    NDIS_STATUS_SUCCESS && n > 0)
 		TRACE2("mac options supported: 0x%x", n);
 
-	transport_header_offset.protocol_type = NDIS_PROTOCOL_ID_TCP_IP;
-	transport_header_offset.header_offset = sizeof(ETH_HLEN);
-	ndis_status = miniport_set_info(wnd, OID_GEN_TRANSPORT_HEADER_OFFSET,
-					&transport_header_offset,
-					sizeof(transport_header_offset));
-	TRACE2("%08X", ndis_status);
+	tx_header_offset = (typeof(tx_header_offset))buf;
+	tx_header_offset->protocol_type = NDIS_PROTOCOL_ID_TCP_IP;
+	tx_header_offset->header_offset = sizeof(ETH_HLEN);
+	status = miniport_set_info(wnd, OID_GEN_TRANSPORT_HEADER_OFFSET,
+				   tx_header_offset, sizeof(*tx_header_offset));
+	TRACE2("%08X", status);
 
 	wnd->tx_ok = 1;
 	if (wnd->physical_medium == NdisPhysicalMediumWirelessLan) {
 		miniport_set_int(wnd, OID_802_11_POWER_MODE, NDIS_POWER_OFF);
-		miniport_set_int(wnd, OID_802_11_NETWORK_TYPE_IN_USE,
-				 Ndis802_11Automode);
-		get_encryption_capa(wnd);
+		get_encryption_capa(wnd, buf, buf_len);
 		TRACE1("capbilities = %ld", wnd->capa.encr);
 		printk(KERN_INFO "%s: encryption modes supported: "
 		       "%s%s%s%s%s%s%s\n", net_dev->name,
@@ -1938,7 +1944,6 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 		       ", WPA2PSK" : "");
 
 		set_infra_mode(wnd, Ndis802_11Infrastructure);
-		set_scan(wnd);
 		set_priv_filter(wnd, Ndis802_11PrivFilterAcceptAll);
 		set_auth_mode(wnd, Ndis802_11AuthModeOpen);
 		set_encr_mode(wnd, Ndis802_11EncryptionDisabled);
@@ -1947,7 +1952,7 @@ static NDIS_STATUS ndis_start_device(struct wrap_ndis_device *wnd)
 	kfree(buf);
 	wrap_procfs_add_ndis_device(wnd);
 	hangcheck_add(wnd);
-	add_stats_timer(wnd);
+	add_iw_stats_timer(wnd);
 	EXIT1(return NDIS_STATUS_SUCCESS);
 
 buffer_pool_err:
@@ -1960,11 +1965,11 @@ packet_pool_err:
 err_register:
 	kfree(buf);
 err_start:
-	ndis_remove_device(wnd);
+	wrap_ndis_remove_device(wnd);
 	EXIT1(return NDIS_STATUS_FAILURE);
 }
 
-static int ndis_remove_device(struct wrap_ndis_device *wnd)
+static int wrap_ndis_remove_device(struct wrap_ndis_device *wnd)
 {
 	s8 tx_pending;
 
@@ -1972,17 +1977,17 @@ static int ndis_remove_device(struct wrap_ndis_device *wnd)
 	memset(&wnd->essid, 0, sizeof(wnd->essid));
 	if (wnd->physical_medium == NdisPhysicalMediumWirelessLan) {
 		up(&wnd->ndis_comm_mutex);
-		miniport_set_info(wnd, OID_802_11_DISASSOCIATE, NULL, 0);
+		disassociate(wnd, 0);
 		down_interruptible(&wnd->ndis_comm_mutex);
 	}
 	set_bit(SHUTDOWN, &wnd->wrap_ndis_pending_work);
-	unregister_netdevice_notifier(&netdev_notifier);
 	wnd->tx_ok = 0;
 	unregister_netdev(wnd->net_dev);
 	netif_carrier_off(wnd->net_dev);
 	/* if device is suspended, but resume failed, tx_ring_mutex
 	 * may already be locked */
 	down_trylock(&wnd->tx_ring_mutex);
+	nt_spin_lock_bh(&wnd->tx_ring_lock);
 	tx_pending = wnd->tx_ring_end - wnd->tx_ring_start;
 	if (tx_pending < 0)
 		tx_pending += TX_RING_SIZE;
@@ -1998,6 +2003,7 @@ static int ndis_remove_device(struct wrap_ndis_device *wnd)
 		wnd->tx_ring_start = (wnd->tx_ring_start + 1) % TX_RING_SIZE;
 		tx_pending--;
 	}
+	nt_spin_unlock_bh(&wnd->tx_ring_lock);
 	up(&wnd->tx_ring_mutex);
 	wrap_procfs_remove_ndis_device(wnd);
 	miniport_halt(wnd);
@@ -2056,7 +2062,6 @@ static wstdcall NTSTATUS NdisAddDevice(struct driver_object *drv_obj,
 	TRACE1("wnd: %p", wnd);
 
 	nmb = ((void *)wnd) + sizeof(*wnd);
-	wnd->nmb = nmb;
 #if defined(DEBUG) && DEBUG >= 6
 	/* poison nmb so if a driver accesses uninitialized pointers, we
 	 * know what it is */
@@ -2064,22 +2069,23 @@ static wstdcall NTSTATUS NdisAddDevice(struct driver_object *drv_obj,
 		((unsigned long *)nmb)[i] = i + 0x8a3fc1;
 #endif
 
+	wnd->nmb = nmb;
 	nmb->wnd = wnd;
 	nmb->pdo = pdo;
 	wd->wnd = wnd;
 	wnd->wd = wd;
-	nmb->filterdbs.eth_db = nmb;
-	nmb->filterdbs.tr_db = nmb;
-	nmb->filterdbs.fddi_db = nmb;
-	nmb->filterdbs.arc_db = nmb;
-
-	KeInitializeSpinLock(&nmb->lock);
-	init_nmb_functions(nmb);
 	wnd->net_dev = net_dev;
-	wnd->ndis_irq = NULL;
+	fdo->reserved = wnd;
+	nmb->fdo = fdo;
+	if (ndis_init_device(wnd)) {
+		IoDeleteDevice(fdo);
+		free_netdev(net_dev);
+		EXIT1(return STATUS_RESOURCES);
+	}
+	nmb->next_device = IoAttachDeviceToDeviceStack(fdo, pdo);
+	nt_spin_lock_init(&wnd->tx_ring_lock);
 	init_MUTEX(&wnd->tx_ring_mutex);
 	init_MUTEX(&wnd->ndis_comm_mutex);
-	init_waitqueue_head(&wnd->ndis_comm_wq);
 	wnd->ndis_comm_done = 0;
 	initialize_work(&wnd->tx_work, tx_worker, wnd);
 	wnd->tx_ring_start = 0;
@@ -2095,33 +2101,24 @@ static wstdcall NTSTATUS NdisAddDevice(struct driver_object *drv_obj,
 	wnd->nick[0] = 0;
 	init_timer(&wnd->hangcheck_timer);
 	wnd->scan_timestamp = 0;
-	init_timer(&wnd->stats_timer);
-	wnd->stats_interval = 10 * HZ;
+	init_timer(&wnd->iw_stats_timer);
+	wnd->iw_stats_interval = 10 * HZ;
 	wnd->wrap_ndis_pending_work = 0;
 	memset(&wnd->essid, 0, sizeof(wnd->essid));
 	memset(&wnd->encr_info, 0, sizeof(wnd->encr_info));
 	wnd->infrastructure_mode = Ndis802_11Infrastructure;
 	initialize_work(&wnd->wrap_ndis_work, wrap_ndis_worker, wnd);
-	wnd->hw_status = 0;
-	if (wd->driver->ndis_driver)
-		wd->driver->ndis_driver->miniport.shutdown = NULL;
-	wnd->stats_enabled = TRUE;
-	InitializeListHead(&wnd->timer_list);
+	wnd->iw_stats_enabled = TRUE;
 
-	fdo->reserved = wnd;
-	nmb->fdo = fdo;
-	nmb->next_device = IoAttachDeviceToDeviceStack(fdo, pdo);
 	TRACE1("nmb: %p, pdo: %p, fdo: %p, attached: %p, next: %p",
 	       nmb, pdo, fdo, fdo->attached, nmb->next_device);
 
 	/* dispatch routines are called as Windows functions */
 	for (i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++)
-		drv_obj->major_func[i] =
-			WIN_FUNC_PTR(IoPassIrpDown,2);
-	drv_obj->major_func[IRP_MJ_PNP] =
-		WIN_FUNC_PTR(NdisDispatchPnp,2);
-	drv_obj->major_func[IRP_MJ_POWER] =
-		WIN_FUNC_PTR(NdisDispatchPower,2);
+		drv_obj->major_func[i] = WIN_FUNC_PTR(IoPassIrpDown,2);
+
+	drv_obj->major_func[IRP_MJ_PNP] = WIN_FUNC_PTR(NdisDispatchPnp,2);
+	drv_obj->major_func[IRP_MJ_POWER] = WIN_FUNC_PTR(NdisDispatchPower,2);
 	drv_obj->major_func[IRP_MJ_INTERNAL_DEVICE_CONTROL] =
 		WIN_FUNC_PTR(NdisDispatchDeviceControl,2);
 //	drv_obj->major_func[IRP_MJ_DEVICE_CONTROL] =
@@ -2135,3 +2132,24 @@ int init_ndis_driver(struct driver_object *drv_obj)
 	drv_obj->drv_ext->add_device = NdisAddDevice;
 	return 0;
 }
+
+int wrapndis_init(void)
+{
+	wrapndis_wq = create_singlethread_workqueue("wrapndis_wq");
+	if (!wrapndis_wq)
+		EXIT1(return -ENOMEM);
+	wrapndis_worker_thread = wrap_worker_init(wrapndis_wq);
+	TRACE1("%p", wrapndis_worker_thread);
+	register_netdevice_notifier(&netdev_notifier);
+	return 0;
+}
+
+void wrapndis_exit(void)
+{
+	unregister_netdevice_notifier(&netdev_notifier);
+	if (wrapndis_wq)
+		destroy_workqueue(wrapndis_wq);
+	TRACE1("%p", wrapndis_worker_thread);
+	if (wrapndis_worker_thread)
+		ObDereferenceObject(wrapndis_worker_thread);
+}
