@@ -19,6 +19,9 @@
 #include "ndis.h"
 #include "pnp.h"
 
+int wrapndis_init(void);
+void wrapndis_exit(void);
+
 NDIS_STATUS miniport_reset(struct wrap_ndis_device *wnd);
 NDIS_STATUS miniport_query_info_needed(struct wrap_ndis_device *wnd,
 				       ndis_oid oid, void *buf,
@@ -36,7 +39,6 @@ void free_tx_packet(struct wrap_ndis_device *wnd, struct ndis_packet *packet,
 int init_ndis_driver(struct driver_object *drv_obj);
 NDIS_STATUS ndis_reinit(struct wrap_ndis_device *wnd);
 
-void get_encryption_capa(struct wrap_ndis_device *wnd);
 void hangcheck_add(struct wrap_ndis_device *wnd);
 void hangcheck_del(struct wrap_ndis_device *wnd);
 
@@ -44,6 +46,6 @@ driver_dispatch_t winNdisDispatchPnp;
 driver_dispatch_t winNdisDispatchPower;
 driver_dispatch_t winNdisDispatchDeviceControl;
 
-struct iw_statistics *get_wireless_stats(struct net_device *dev);
+struct iw_statistics *get_iw_stats(struct net_device *dev);
 
 #endif
