@@ -50,7 +50,7 @@ static int debug;
 #endif
 
 #define LOG_MSG(where, fmt, ...)					\
-	syslog(LOG_KERN | (where), "%s: %s(%d): " fmt "\n",		\
+	syslog(LOG_KERN | where, "%s: %s(%d): " fmt "\n",		\
 	       PROG_NAME, __FUNCTION__, __LINE__ , ## __VA_ARGS__)
 #define ERROR(fmt, ...) LOG_MSG(LOG_INFO, fmt, ## __VA_ARGS__)
 #define INFO(fmt, ...) LOG_MSG(LOG_INFO, fmt, ## __VA_ARGS__)
@@ -152,7 +152,7 @@ static int read_conf_file(char *conf_file_name, struct load_driver *driver)
 	char setting_name[MAX_SETTING_NAME_LEN];
 	char setting_value[MAX_SETTING_VALUE_LEN];
 	int ret, nr_settings;
-	int i, vendor, device, subvendor, subdevice, bus;
+	int vendor, device, subvendor, subdevice, bus;
 
 	if (lstat(conf_file_name, &statbuf)) {
 		ERROR("unable to open config file %s: %s",
@@ -168,8 +168,7 @@ static int read_conf_file(char *conf_file_name, struct load_driver *driver)
 			  &bus) == 5) {
 		DBG("bus: %X", bus);
 	} else {
-		ERROR("unable to parse conf file name %s (%d)",
-		      conf_file_name, i);
+		ERROR("unable to parse conf file name %s", conf_file_name);
 		return -EINVAL;
 	}
 
@@ -437,15 +436,14 @@ static int load_device(int ioctl_device, int vendor, int device,
 		if (dirent->d_name[0] == '.')
 			continue;
 
-		if (!get_device(dirent->d_name, vendor, device,
-				subvendor, subdevice, bus, &load_device))
+		if (!get_device(dirent->d_name, vendor, device, subvendor,
+				subdevice, bus, &load_device))
 			break;
 	}
 	closedir(dir);
 
 	DBG("%04x, %04x, %04x, %04x", load_device.vendor,
-	    load_device.device, load_device.subvendor,
-	    load_device.subdevice);
+	    load_device.device, load_device.subvendor, load_device.subdevice);
 	res = ioctl(ioctl_device, WRAP_IOCTL_LOAD_DEVICE, &load_device);
 	DBG("res: %d", res);
 
